module Lexer(
  Token (..),
  lexStr
  )where

import Data.Char

data Token = TokLPar
           | TokRPar
           | TokQuote
           | TokNumber Double
           | TokStr String
           | TokCharacter Char
           | TokSymbol String
           | TokErr String
  deriving (Eq, Show)

isWhitespace :: Char -> Bool
isWhitespace x = x == ' ' || x == '\t' || x == '\n' || x == '\r'

data StrState = StrStart
              | StrInside
              | StrEscape

identifySymbol :: String -> (String, String)
identifySymbol "" = ("", "")
identifySymbol (x:xs) | isWhitespace x = (x:xs, "")
                      | otherwise = case identifySymbol xs of
                                      (ys, zs) -> (ys, x:zs)

identifyString :: String -> StrState -> Maybe (String, String)
identifyString ('"':xs) StrStart = case identifyString xs StrInside of
                                     Nothing -> Nothing
                                     y -> y
identifyString _ StrStart = Nothing -- should never occur unless not called on a string
identifyString "" _ = Nothing -- early end of file
identifyString ('\\':xs) StrInside = identifyString xs StrEscape
identifyString ('"':xs) StrInside = Just (xs, "")
identifyString ('n':xs) StrEscape = case identifyString xs StrInside of
  Just (ys, b) -> Just (ys, '\n':b)
  y -> y
identifyString ('\\':xs) StrEscape = case identifyString xs StrInside of
  Just (ys, b) -> Just (ys, '\\':b)
  y -> y
identifyString ('/':xs) StrEscape = case identifyString xs StrInside of
  Just (ys, b) -> Just (ys, '/':b)
  y -> y
identifyString ('"':xs) StrEscape = case identifyString xs StrInside of
  Just (ys, b) -> Just (ys, '"':b)
  y -> y
identifyString ('b':xs) StrEscape = case identifyString xs StrInside of
  Just (ys, b) -> Just (ys, '\b':b)
  y -> y
identifyString ('f':xs) StrEscape = case identifyString xs StrInside of
  Just (ys, b) -> Just (ys, '\f':b)
  y -> y
identifyString ('r':xs) StrEscape = case identifyString xs StrInside of
  Just (ys, b) -> Just (ys, '\r':b)
  y -> y
identifyString ('t':xs) StrEscape = case identifyString xs StrInside of
  Just (ys, b) -> Just (ys, '\t':b)
  y -> y
identifyString ('u':w:x:y:z:xs) StrEscape
  | isHexDigit w &&
    isHexDigit x &&
    isHexDigit y &&
    isHexDigit z =
      case identifyString xs StrInside of
        Just (ys, b) -> Just (ys, chr (read ('0':'x':w:x:y:z:"")):b)
        a -> a
identifyString _ StrEscape = Nothing
identifyString (x:xs) StrInside = case identifyString xs StrInside of
     Just (ys, b) -> Just (ys, x:b)
     y -> y

data NumState = NumStart
              | NumNeg
              | NumIntegralComponent
              | NumFractionalComponent
              | NumExponentialComponent

identifyNum :: String -> NumState -> Maybe (String, String)
identifyNum "" _ = Nothing
identifyNum (x:"") _ | isDigit x = Just ("", x:"")
identifyNum (x:y:xs) _ | isDigit x && isWhitespace y = Just (y:xs, x:"")
identifyNum ('-':xs) NumStart = case identifyNum xs NumNeg of
                                  Just (ys, zs) -> Just (ys, '-':zs)
                                  y -> y
identifyNum ('0':'.':x:xs) NumStart | isDigit x = case identifyNum (x:xs) NumFractionalComponent of
                                                    Just (ys, zs) -> Just (ys, '0':'.':zs)
                                                    y -> y
identifyNum (x:xs) NumStart | x == '0' = Nothing
                            | isDigit x = identifyNum (x:xs) NumIntegralComponent
                            | otherwise = Nothing
identifyNum ('0':'.':x:xs) NumNeg | isDigit x = case identifyNum (x:xs) NumFractionalComponent of
                                                  Just (ys, zs) -> Just (ys, '0':'.':zs)
                                                  y -> y
identifyNum (x:xs) NumNeg | x == '0' = Nothing
                          | isDigit x = identifyNum (x:xs) NumIntegralComponent
                          | otherwise = Nothing
identifyNum ('.':x:xs) NumIntegralComponent
  | isDigit x = case identifyNum (x:xs) NumFractionalComponent of
                  Just (ys, zs) -> Just (ys, '.':zs)
                  y -> y
  | otherwise = Nothing
identifyNum (w:x:y:xs) NumIntegralComponent | (w == 'e' || w == 'E') &&
                                              (x == '+' || x == '-') &&
                                                isDigit y =
                                                  case identifyNum (y:xs) NumExponentialComponent of
                                                    Just (ys, zs) -> Just (ys, w:x:zs)
                                                    z -> z
identifyNum (x:y:xs) NumIntegralComponent | (x == 'e' || x == 'E') &&
                                            isDigit y =
                                              case identifyNum (y:xs) NumExponentialComponent of
                                                Just (ys, zs) -> Just (ys, x:zs)
                                                z -> z
identifyNum (x:xs) NumIntegralComponent | isDigit x = case identifyNum xs NumIntegralComponent of
                                                        Just (ys, zs) -> Just (ys, x:zs)
                                                        y -> y
                                        | otherwise = Nothing
identifyNum (w:x:y:xs) NumFractionalComponent | (w == 'e' || w == 'E') &&
                                                (x == '+' || x == '-') &&
                                                isDigit y =
                                                  case identifyNum (y:xs) NumExponentialComponent of
                                                    Just (ys, zs) -> Just (ys, w:x:zs)
                                                    z -> z
identifyNum (x:y:xs) NumFractionalComponent | (x == 'e' || x == 'E') &&
                                              isDigit y =
                                                case identifyNum (y:xs) NumExponentialComponent of
                                                  Just (ys, zs) -> Just (ys, x:zs)
                                                  z -> z
identifyNum (x:xs) NumFractionalComponent | isDigit x = case identifyNum xs NumFractionalComponent of
                                                          Just (ys, zs) -> Just (ys, x:zs)
                                                          y -> y
                                          | otherwise = Nothing
identifyNum (x:xs) NumExponentialComponent | isDigit x = case identifyNum xs NumExponentialComponent of
                                                           Just (ys, zs) -> Just (ys, x:zs)
                                                           y -> y
                                           | otherwise = Nothing

lexStr :: String -> [Token]
lexStr "" = []
lexStr ('(':xs) = TokLPar : lexStr xs
lexStr (')':xs) = TokRPar : lexStr xs
lexStr ('"':xs) = case identifyString ('"':xs) StrStart
  of Nothing -> case identifySymbol ('"':xs) of (rest, symb) -> TokSymbol symb : lexStr rest
     Just (ys, b) -> TokStr b:lexStr ys
lexStr ('\'':"") = [TokSymbol "'"]
lexStr ('\'':'(':xs) = TokQuote : lexStr ('(':xs)
lexStr ('\'':x:'\'':xs) = TokCharacter x : lexStr xs
lexStr ('\'':x:xs) | isWhitespace x = TokSymbol "'" : lexStr (x:xs)
                   | otherwise = case identifySymbol (x:xs) of
                                   (rest, symb) -> TokQuote : TokSymbol symb : lexStr rest
lexStr (' ':xs) = lexStr xs
lexStr ('\t':xs) = lexStr xs
lexStr ('\n':xs) = lexStr xs
lexStr ('\r':xs) = lexStr xs
lexStr (x:xs) | isDigit x ||
                x == '-' =
                  case identifyNum (x:xs) NumStart of
                    Nothing -> case identifySymbol (x:xs) of (rest, symb) -> TokSymbol symb : lexStr rest
                    Just (ys, b) -> TokNumber (read b) : lexStr ys
lexStr xs = case identifySymbol xs of (ys, zs) -> TokSymbol zs : lexStr ys
